@@ -16,6 +16,7 @@ package cmd
 
 import (
 	"bufio"
+	"fmt"
 	"strings"
 
 	"github.com/spf13/cobra"
@@ -31,39 +32,59 @@ import (
 // The user can then scroll up to get a more in-depth explanation. But is
 // this how users use it?
 
-// newHelpCmd is largely borrowed from cobra,
+// newHelpCmd is similar to the built-in help command from cobra,
 // but knows how to load custom commands in `cue help cmd`.
 func newHelpCmd(c *Command) *cobra.Command {
 	cmd := &cobra.Command{
-		Use:   "help [command]",
-		Short: "Help about any command",
-		Long: `Help provides help for any command in the application.
-Simply type ` + c.Name() + ` help [path to command] for full details.`,
+		Use:    "help [command]",
+		Short:  "show help text for a command or topic",
+		Hidden: true, // not shown as a runnable command in `cue help`
 		Run: func(_ *cobra.Command, args []string) {
-			cmd, _, e := c.Root().Find(args)
-			if len(args) > 0 && args[0] == "cmd" {
+			findCmd := func() (*cobra.Command, bool) {
+				cmd, rest, err := c.Root().Find(args)
+				found := cmd != nil && err == nil && len(rest) == 0
+				return cmd, found
+			}
+			cmd, found := findCmd()
+			// Treat `cue help help` as an unknown help topic; just use `cue help`.
+			if found && cmd.Name() == "help" {
+				cmd, found = nil, false
+			}
+			isCmd := len(args) > 0 && args[0] == "cmd"
+			if isCmd {
 				// args is one of:
 				//
 				//	["cmd"]
 				//	["cmd", "mycmd"]
 				//	["cmd", "mycmd", "./mypkg"]
 				//
-				// We want to skip the first two arguments in pkgArgs.
-				pkgArgs := args[1:]
-				if len(pkgArgs) > 0 {
-					pkgArgs = pkgArgs[1:]
+				// In the third case, we want to load ["./mypkg"]
+				// and we want to look up the help for ["cmd", "mycmd"].
+				var pkgArgs []string
+				if len(args) > 2 {
+					pkgArgs = args[2:]
+					args = args[:2]
 				}
 
 				tools, err := buildTools(c, pkgArgs)
 				if err == nil {
 					addCustomCommands(c, cmd, commandSection, tools)
 					// For the sake of `cue help cmd mycmd`, find the command again.
-					cmd, _, e = c.Root().Find(args)
+					cmd, found = findCmd()
 				}
 			}
-			if cmd == nil || e != nil {
-				c.Printf("Unknown help topic %#q\n", args)
-				cobra.CheckErr(c.Root().Usage())
+			if !found {
+				if isCmd {
+					// Note that for args ["cmd", "mycmd", "./mypkg"] we only want "mycmd".
+					fmt.Fprintf(c.Stderr(), "Unknown cmd command: %s\n", args[1])
+				} else {
+					fmt.Fprintf(c.Stderr(), "Unknown help topic: %s\n", strings.Join(args, " "))
+				}
+				if cmd == nil {
+					cobra.CheckErr(c.Root().Usage())
+				} else {
+					cobra.CheckErr(cmd.Usage())
+				}
 			} else {
 				cobra.CheckErr(cmd.Help())
 			}
@@ -72,20 +93,44 @@ Simply type ` + c.Name() + ` help [path to command] for full details.`,
 	return cmd
 }
 
-// TODO(mvdan): having the help topics as top-level commands means that `cue topic`
-// is taken and works as well as `cue help topic`, which is unnecessary.
-// Consider removing support for the short form at some point.
+// We use a custom Cobra help text template for the sake of being a bit more
+// in control of its formatting, as well as the contents of the root `cue help`.
+var helpTemplate = `
+{{- if not .HasParent}}{{/* Special template for the root help. */ -}}
+CUE makes it easy to validate data, write schemas,
+and ensure configurations align with policies.
 
-func newHelpTopics(c *Command) []*cobra.Command {
-	return []*cobra.Command{
-		inputsHelp,
-		environmentHelp,
-		flagsHelp,
-		filetypeHelp,
-		injectHelp,
-		commandsHelp,
-		registryConfigHelp,
-	}
+CUE works with a wide range of tools and formats that you're already using
+such as Go, JSON, YAML, OpenAPI, and JSON Schema.
+
+For more information and documentation, see: https://cuelang.org
+
+Available Commands:{{range .Commands}}{{if .IsAvailableCommand}}
+  {{rpad .Name .NamePadding }} {{.Short}}{{end}}{{end}}
+
+Use "{{.CommandPath}} help [command]" for more information about a command.
+
+{{/* We want to show additional help topics from the root command's help text
+     even though they are only reachable via 'cue help'. */ -}}
+Additional help topics:{{range .Commands}}{{if eq .Name "help"}}{{range .Commands}}
+  {{rpad .CommandPath .CommandPathPadding}} {{.Short}}{{end}}{{end}}{{end}}
+{{else}}{{/* Subcommands use a fairly standard template. */ -}}
+{{with (or .Long .Short)}}{{. | trimTrailingWhitespaces}}
+
+{{end}}{{if or .Runnable .HasSubCommands}}{{.UsageString}}{{end}}
+{{- end -}}
+`
+
+var helpTopics = []*cobra.Command{
+	commandsHelp,
+	embedHelp,
+	environmentHelp,
+	filetypeHelp,
+	flagsHelp,
+	injectHelp,
+	inputsHelp,
+	modulesHelp,
+	registryConfigHelp,
 }
 
 var inputsHelp = &cobra.Command{
@@ -101,21 +146,43 @@ the package in the current directory, provided there is a single
 named package in this directory.
 
 CUE packages are specified as an import path. An import path
-that is a rooted path --one that begins with a "." or ".."
-element-- is interpreted as a file system path and denotes the
+that begins with a "." or ".." element is a relative import path,
+is interpreted as a file system path, and denotes the
 package instance in that directory.
 
-Otherwise, the import path P denotes and external package found
-in cue.mod/{pkg|gen|usr}/P.
+Otherwise, the import path P is absolute and denotes a package that
+may be external, usually found in an external registry. It may also
+refer to a package in cue.mod/{pkg|gen|usr}/P - this is legacy
+behavior that might be removed at a later date.
 
-An import path may contain one or more "..." to match any
-subdirectory: pkg/... matches all packages below pkg, including
-pkg itself, while foo/.../bar matches all directories named bar
-within foo. In all cases, directories containing cue.mod
-directories are excluded from the result.
+An absolute import path is of the form P or P@vN where vN is the major
+version of the module containing the package. An import path is mapped
+to a registry location by consulting cue.mod/module.cue and the
+registry configuration (see "cue help modules" and "cue help
+registryconfig" for more details).
+
+An import may contain a ":name" suffix to indicate a specific
+package to load within a given path. It is necessary to specify this
+when there is more than one package at the path.
+
+For example, the following import path specifies a package
+named "other" in some external module with major version v1.
+
+	foo.example/bar/baz@v1:other
+
+A local import path may contain one or more "..." to match any
+subdirectory: pkg/... matches all packages below pkg, including pkg
+itself, while foo/.../bar matches all directories named bar within
+foo. In all cases, directories containing cue.mod directories are
+excluded from the result. "..." is not currently supported in external
+import paths.
 
 Directory and file names that begin with "." or "_" are ignored,
-unless explicitly listed as inputs.
+unless explicitly listed as inputs. File with names ending "_tool.cue"
+are ignored unless running "cue cmd" and they are in packages
+explicitly mentioned on the command line. Files with names ending
+"_test.cue" are ignored for the time being; they are reserved for
+future testing functionality.
 
 A package may also be specified as a list of .cue files.
 The special symbol '-' denotes stdin or stdout and defaults to
@@ -154,54 +221,138 @@ $ cue export data.json schema: schema.json
 `,
 }
 
+var embedHelp = &cobra.Command{
+	Use:   "embed",
+	Short: "file embedding",
+	Long: `
+The @embed attribute can be used to load files as part of CUE evaluation.
+Embedding must be enabled by a file-level @extern(embed) attribute.
+
+Embed a single file as follows, which must exist when evaluating:
+
+	@extern(embed)
+
+	package foo
+
+	a: _ @embed(file=foo.json)
+
+Embed many files of the same type with a glob pattern,
+which inserts one field per file matched:
+
+	b: _ @embed(glob=images/*.*, type=binary)
+	b: [string]: bytes
+
+Note that "**" glob patterns are not supported at this time.
+
+If the file extension in "file" or "glob" does not imply a file type,
+it must be specified with the "type" encoding as shown above.
+See the "filetypes" help topic for more.
+
+Files containing multiple documents cannot be embedded directly.
+For NDJSON or multi-document YAML files, embed as type=text
+and use APIs like yaml.Extract to decode as a list.
+
+Note that embedding cue files is not supported at this time.
+
+Note: support for embed is experimental, and is enabled via
+CUE_EXPERIMENT=embed.
+
+For more details and discussion, see the proposal linked from
+https://cuelang.org/discussion/3264.
+`[1:],
+}
+
 var environmentHelp = &cobra.Command{
 	Use:   "environment",
 	Short: "environment variables",
-	Long: `The cue command consults environment variables for configuration.
-If an environment variable is unset or empty, it uses a sensible default
-setting.
+	Long: `
+The cue command consults environment variables for configuration.
+If an environment variable is unset or empty, sensible default setting is used.
 
 	CUE_CACHE_DIR
-		The directory where the cue command keeps a cache of files to be
-		reused. This defaults to a directory for user-specific temporary
-		cache data, such as:
+		A directory structure to hold files which are reused between operations.
+		It defaults to a directory for user-specific temporary data, such as:
 
-			"$XDG_CACHE_HOME/cue" or "$HOME/.cache/cue" on Linux
-			"$HOME/Library/Caches/cue" on MacOS
-			"%LocalAppData%/cue" on Windows
+		- "$XDG_CACHE_HOME/cue" or "$HOME/.cache/cue" on Linux
+		- "$HOME/Library/Caches/cue" on MacOS
+		- "%LocalAppData%/cue" on Windows
+
+		Its contents are generally read-only and include:
+
+		- mod/download for modules fetched from registries
+		- mod/extract for extracted module archives
 
 	CUE_CONFIG_DIR
-		The directory where the cue command keeps configuration and
-		state files. This defaults to a directory for user-specific
-		configuration data, such as:
+		A directory to hold configuration and long-lived state files.
+		It defaults to a directory for user-specific persistent data, such as:
 
-			"$XDG_CONFIG_HOME/cue" or "$HOME/.config/cue" on Linux
-			"$HOME/Library/Application Support/cue" on MacOS
-			"%AppData%/cue" on Windows
-
-	CUE_DEBUG
-		Comma-separated list of debug flags to enable. Current possible
-		values are:
-			http
-				Print log messages in JSON format, one per line showing
-				HTTP requests and responses made when interacting with
-				module registries.
-
-	CUE_EXPERIMENT
-		Comma-separated list of experiments to enable. The
-		list of available experiments may change arbitrarily over time.
-		See exp.go in package cuelang.org/go/internal/cueexperiment for
-		currently valid values.
-
-		Warning: This variable is provided for the development and
-		testing of the cue commands itself. Use beyond that purpose is
-		unsupported.
+		- "$XDG_CONFIG_HOME/cue" or "$HOME/.config/cue" on Linux
+		- "$HOME/Library/Application Support/cue" on MacOS
+		- "%AppData%/cue" on Windows
 
 	CUE_REGISTRY
+		The configuration to use when downloading and publishing modules.
+		See "cue help registryconfig" for details.
 
-		This variable specifies which registry or registries to use for
-		downloading and publishing modules. See "cue help registryconfig"
-		for details.
+	CUE_EXPERIMENT
+		Comma-separated list of experiment flags to enable or disable:
+
+		evalv3
+			Enable the new CUE evaluator, addressing performance issues
+			and bringing a better disjunction algorithm.
+		embed
+			Enable support for embedded data files as described in
+			https://cuelang.org/discussion/3264.
+		decodeint64
+			Tweak cue.Value.Decode to choose "int64" rather than "int"
+			as the default Go type for CUE integer values.
+
+	CUE_DEBUG
+		Comma-separated list of debug flags to enable or disable, such as:
+
+		http
+			Log a JSON message per HTTP request and response made
+			when interacting with module registries.
+
+CUE_EXPERIMENT and CUE_DEBUG are comma-separated lists of key-value strings,
+where the value is a boolean "true" or "1" if omitted. For example:
+
+	CUE_EXPERIMENT=toenable,todisable=0
+`[1:],
+}
+
+var modulesHelp = &cobra.Command{
+	Use:   "modules",
+	Short: "module support",
+	Long: `Modules are how CUE publishes packages and manages dependencies.
+
+A module is a collection of packages that are released, versioned, and
+distributed together. Modules are downloaded from OCI-compliant
+artifact registries.
+
+The Central Registry for CUE modules (` + modconfig.DefaultRegistry + `) is
+used as the source for external modules by default.
+
+For a detailed reference on modules:
+
+	https://cuelang.org/docs/reference/modules/
+
+For information on commands that interact with modules:
+
+    cue help mod
+
+For tutorials on how to use the Central Registry, see:
+
+	https://cuelang.org/docs/tutorial/working-with-the-central-registry/
+	https://cuelang.org/docs/tutorial/publishing-modules-to-the-central-registry/
+
+For a tutorial on how to work with a custom OCI registry for CUE modules:
+
+	https://cuelang.org/docs/tutorial/working-with-a-custom-module-registry/
+
+For information on how to specify the CUE registry:
+
+	cue help registryconfig
 `,
 }
 
@@ -215,8 +366,8 @@ Given a particular registry configuration, it's possible to work what registry a
 given module is stored in, and what repository within that registry, without
 making any network calls.
 
-If no explicit registry configuration is present, a default central registry
-(` + modconfig.DefaultRegistry + `) will be used for all modules.
+If no explicit registry configuration is present, a default Central Registry
+(` + modconfig.DefaultRegistry + `) is used for all modules.
 
 The simplest way of specifying a registry configuration is to set $CUE_REGISTRY
 to the hostname of that registry.
@@ -227,6 +378,9 @@ Examples:
 	CUE_REGISTRY=localhost:5000
 	CUE_REGISTRY='[::1]:5000'
 
+The special name "none" can be used to indicate that no registry
+should be used.
+
 If a path is present too, all modules will be stored under that path.
 
 For example:
@@ -263,6 +417,18 @@ Note that the syntax above implies that the ordering of the elements in
 CUE_REGISTRY isn't important because the resolution algorithm is
 order-independent.
 
+To specify that no registry should be used for a given module prefix,
+the special name "none" can be used.
+
+For example:
+
+	CUE_REGISTRY='foo.example/bar=none,myregistry.example'
+
+In the above example, any attempt to fetch a module under
+"foo.example/bar" will result in a failure. Note that this will not
+take effect if the module is already present in the on-disk cache,
+which is consulted before looking at CUE_REGISTRY.
+
 
 Customizing Name Resolution
 
@@ -376,7 +542,8 @@ var filetypeHelp = &cobra.Command{
     cue         .cue            CUE source files.
     json        .json           JSON files.
     yaml        .yaml/.yml      YAML files.
-    jsonl       .jsonl/.ldjson  Line-separated JSON values.
+    toml        .toml           TOML files
+    jsonl       .jsonl/.ndjson  Line-separated JSON values.
     jsonschema                  JSON Schema.
     openapi                     OpenAPI schema.
 	pb                          Use Protobuf mappings (e.g. json+pb)
@@ -430,6 +597,14 @@ defaults as requested.
     graph       Like data, but allow references.
     schema      Export data and definitions.
 
+The following tags are only valid in combination with other tags,
+and influence the functioning of the codec. The tag they are
+valid with is mentioned in parentheses at the end.
+
+    strictFeatures	report errors for lossy mappings. (jsonschema)
+    strictKeywords	report errors for unknown keywords (jsonschema)
+    strict			report errors for either of the above (jsonschema)
+
 Many commands also support the --out and --outfile/-o flags.
 The --out flag specifies the output type using a qualifier
 (without the ':'). The -o flag specifies an output file
@@ -470,7 +645,8 @@ Injecting files
 A "build" attribute defines a boolean expression that causes a file
 to only be included in a build if its expression evaluates to true.
 There may only be a single @if attribute per file and it must
-appear before a package clause.
+appear before a package clause, or before any CUE declarations
+if there is no package clause.
 
 The expression is a subset of CUE consisting only of identifiers
 and the operators &&, ||, !, where identifiers refer to tags
@@ -484,6 +660,19 @@ if the user includes the flag "-t prod" on the command line.
 
    package foo
 
+Ignoring files
+
+An "ignore" attribute causes a file to be unconditionally excluded
+from a build. The @ignore attribute must appear before a package
+clause or before any other CUE syntax if there is no package clause.
+
+For example:
+
+    @ignore()
+
+    // This file will be excluded for all purposes.
+
+    package foo
 
 Injecting values
 
@@ -563,15 +752,15 @@ The following variables are supported:
 var commandsHelp = &cobra.Command{
 	Use:   "commands",
 	Short: "user-defined commands",
-	Long: `Commands define actions on instances. For example, they may
-specify how to upload a configuration to Kubernetes. Commands are
-defined directly in tool files, which are regular CUE files
-within the same package with a filename ending in _tool.cue.
-These are typically defined at the module root so that they apply
-to all instances.
+	Long: `Workflow commands define actions on instances. For example, they
+may specify how to upload a configuration to Kubernetes. Workflow
+commands are defined directly in tool files, which are regular
+CUE files within the same package with a filename ending in
+_tool.cue.  These are typically defined at the module root so
+that they apply to all instances.
 
-Each command consists of one or more tasks. A task may, for
-example, load or write a file, consult a user on the command
+Each workflow command consists of one or more tasks. A task may,
+for example, load or write a file, consult a user on the command
 line, fetch a web page, and so on. Each task has inputs and
 outputs. Outputs are typically filled out by the task
 implementation as the task completes.
@@ -587,12 +776,10 @@ Available tasks can be found in the package documentation at
 
 	https://pkg.go.dev/cuelang.org/go/pkg/tool?tab=subdirectories
 
-More on tasks can be found in the commands help topic.
-
 Examples:
 
-In this simple example, we define a command called "hello",
-which declares a single task called "print" which uses
+In this simple example, we define a workflow command called
+"hello", which declares a single task called "print" which uses
 "tool/exec.Run" to execute a shell command that echos output to
 the terminal:
 
@@ -612,7 +799,7 @@ the terminal:
 	}
 	EOF
 
-We run the "hello" command like this:
+We run the "hello" workflow command like this:
 
 	$ cue cmd hello
 	Hello World! Welcome to Amsterdam.
@@ -621,12 +808,13 @@ We run the "hello" command like this:
 	Hello Jan! Welcome to Amsterdam.
 
 
-In this example we declare the "prompted" command which has four
-tasks. The first task prompts the user for a string input. The
-second task depends on the first, and echos the response back to
-the user with a friendly message. The third task pipes the output
-from the second to a file. The fourth task pipes the output from
-the second to standard output (i.e. it echos it again).
+In this example we declare the "prompted" workflow command which
+has four tasks. The first task prompts the user for a string
+input. The second task depends on the first, and echos the
+response back to the user with a friendly message. The third task
+pipes the output from the second to a file. The fourth task pipes
+the output from the second to standard output (i.e. it echos it
+again).
 
 	package foo
 
@@ -743,5 +931,3 @@ func indent(s string) string {
 
 // TODO: filetypes:
 // - binpb
-
-// TODO: cue.mod help topic
