@@ -55,6 +55,9 @@ import (
 
 type runFunction func(cmd *Command, args []string) error
 
+// wasmInterp is set when the cuewasm build tag is enbabled.
+var wasmInterp cuecontext.ExternInterpreter
+
 func statsEncoder(cmd *Command) (*encoding.Encoder, error) {
 	file := os.Getenv("CUE_STATS_FILE")
 	if file == "" {
@@ -105,6 +108,14 @@ func mkRunE(c *Command, f runFunction) func(*cobra.Command, []string) error {
 		if err := cuedebug.Init(); err != nil {
 			return err
 		}
+		var opts []cuecontext.Option
+		if wasmInterp != nil {
+			opts = append(opts, cuecontext.Interpreter(wasmInterp))
+		}
+		if cueexperiment.Flags.Embed {
+			opts = append(opts, cuecontext.Interpreter(embed.New()))
+		}
+		c.ctx = cuecontext.New(opts...)
 		// Some init work, such as in internal/filetypes, evaluates CUE by design.
 		// We don't want that work to count towards $CUE_STATS.
 		adt.ResetStats()
@@ -182,39 +193,9 @@ func mkRunE(c *Command, f runFunction) func(*cobra.Command, []string) error {
 // The returned error is always nil, and is a historical artifact.
 func New(args []string) (*Command, error) {
 	cmd := &cobra.Command{
-		Use:   "cue",
+		Use: "cue",
+		// TODO: the short help text below seems to refer to `cue cmd`, like helpTemplate.
 		Short: "cue emits configuration files to user-defined commands.",
-		// TODO(mvdan): the text below should not jump straight into `cue cmd`,
-		// and instead give a high level overview of CUE and its commands.
-		// TODO(mvdan): cobra's help text template ends with
-		//     Use "cue [command] --help" for more information about a command.
-		// which isn't great; use our own template to say
-		//     Use "cue help [command]" for more information about a command.
-		// TODO(mvdan): cobra's help text template suggests help topics
-		// such as `cue inputs`; swap with `cue help inputs`.
-		Long: `cue evaluates CUE files, an extension of JSON, and sends them
-to user-defined commands for processing.
-
-Commands are defined in CUE as follows:
-
-	import "tool/exec"
-	command: deploy: {
-		exec.Run
-		cmd:   "kubectl"
-		args:  ["-f", "deploy"]
-		in:    json.Encode(userValue) // encode the emitted configuration.
-	}
-
-cue can also combine the results of http or grpc request with the input
-configuration for further processing. For more information on defining commands
-run 'cue help cmd' or go to cuelang.org/pkg/cmd.
-
-For more information on writing CUE configuration files see cuelang.org.`,
-
-		// ArbitraryArgs allows us to forward the top-level RunE to cmdCmd.RunE,
-		// which supports `cue mycmd` as a short-cut for `cue cmd mycmd`.
-		// Without ArbitraryArgs, cobra fails with "unknown command" errors.
-		Args: cobra.ArbitraryArgs,
 
 		// We print errors ourselves in Main, which allows for ErrPrintedError.
 		// Similarly, we don't want to print the entire help text on any error.
@@ -224,23 +205,36 @@ For more information on writing CUE configuration files see cuelang.org.`,
 
 		// We currently support top-level custom commands like `cue mycmd` as an alias
 		// for `cue cmd mycmd`, so any sub-command suggestions might be confusing.
+		// TODO(mvdan): remove this and test it, as we no longer support `cue mycmd`.
 		DisableSuggestions: true,
 	}
 
-	embedding := cuecontext.Interpreter(embed.New())
-	rootContextOptions = append(rootContextOptions, embedding)
-
 	c := &Command{
 		Command: cmd,
 		root:    cmd,
-		ctx:     cuecontext.New(rootContextOptions...),
 	}
+	c.cmdCmd = newCmdCmd(c)
 
-	cmdCmd := newCmdCmd(c)
-	c.cmd = cmdCmd
+	addGlobalFlags(cmd.PersistentFlags())
 
-	subCommands := []*cobra.Command{
-		cmdCmd,
+	// Cobra's --help flag shows up in help text by default, which is unnecessary.
+	cmd.InitDefaultHelpFlag()
+	cmd.Flag("help").Hidden = true
+
+	// "help" is treated as a special command by cobra.
+	// We use our own template to be more in control of the structure of `cue help`.
+	// Note that we need to add helpCmd as a subcommand first, for cobra to work out
+	// the proper help text paddings for additional help topics.
+	helpCmd := newHelpCmd(c)
+	cmd.AddCommand(helpCmd)
+	for _, sub := range helpTopics {
+		helpCmd.AddCommand(sub)
+	}
+	cmd.SetHelpCommand(helpCmd)
+	cmd.SetHelpTemplate(helpTemplate)
+
+	for _, sub := range []*cobra.Command{
+		c.cmdCmd,
 		newCompletionCmd(c),
 		newEvalCmd(c),
 		newDefCmd(c),
@@ -249,6 +243,7 @@ For more information on writing CUE configuration files see cuelang.org.`,
 		newFmtCmd(c),
 		newGetCmd(c),
 		newImportCmd(c),
+		newLoginCmd(c),
 		newModCmd(c),
 		newTrimCmd(c),
 		newVersionCmd(c),
@@ -256,37 +251,15 @@ For more information on writing CUE configuration files see cuelang.org.`,
 
 		// Hidden
 		newAddCmd(c),
-		newLoginCmd(c),
-	}
-	subCommands = append(subCommands, helpTopics...)
-
-	addGlobalFlags(cmd.PersistentFlags())
-	// We add the injection flags to the root command for the sake of the short form "cue -t foo=bar mycmd".
-	// Note that they are not persistent, so that they aren't inherited by sub-commands like "cue fmt".
-	addInjectionFlags(cmd.Flags(), false, true)
-
-	for _, sub := range subCommands {
+		newLSPCmd(c),
+	} {
 		cmd.AddCommand(sub)
 	}
 
-	// Cobra's --help flag shows up in help text by default, which is unnecessary.
-	cmd.InitDefaultHelpFlag()
-	cmd.Flag("help").Hidden = true
-
-	// "help" is treated as a special command by cobra.
-	// TODO(mvdan): hide this command; `cue help` showing itself in the list of commands
-	// is not particularly helpful, and we already mention it as the way to get help.
-	cmd.SetHelpCommand(newHelpCmd(c))
-
-	// For `cue mycmd` to be a shortcut for `cue cmd mycmd`.
-	cmd.RunE = cmdCmd.RunE
-
 	cmd.SetArgs(args)
 	return c, nil
 }
 
-var rootContextOptions []cuecontext.Option
-
 // rootWorkingDir avoids repeated calls to [os.Getwd] in cmd/cue.
 // If we can't figure out the current directory, something is very wrong,
 // and there's no point in continuing to run a command.
@@ -320,8 +293,8 @@ type Command struct {
 
 	root *cobra.Command
 
-	// Subcommands
-	cmd *cobra.Command
+	// _tool.cue subcommands.
+	cmdCmd *cobra.Command
 
 	ctx *cue.Context
 
