@@ -14,7 +14,7 @@
 
 // Package ast declares the types used to represent syntax trees for CUE
 // packages.
-package ast // import "cuelang.org/go/cue/ast"
+package ast
 
 import (
 	"fmt"
@@ -112,7 +112,7 @@ type decl struct{}
 
 func (decl) declNode() {}
 
-// A Label is any production that can be used as a LHS label.
+// A Label is any production that can be used as an LHS label.
 type Label interface {
 	Node
 	labelNode()
@@ -176,10 +176,10 @@ func (c *comments) SetComments(cgs []*CommentGroup) {
 	*c.groups = cgs
 }
 
-// A Comment node represents a single //-style or /*-style comment.
+// A Comment node represents a single //-style comment.
 type Comment struct {
 	Slash token.Pos // position of "/" starting the comment
-	Text  string    // comment text (excluding '\n' for //-style comments)
+	Text  string    // comment text excluding '\n'
 }
 
 func (c *Comment) Comments() []*CommentGroup { return nil }
@@ -226,7 +226,7 @@ func stripTrailingWhitespace(s string) string {
 }
 
 // Text returns the text of the comment.
-// Comment markers (//, /*, and */), the first space of a line comment, and
+// Comment markers ("//"), the first space of a line comment, and
 // leading and trailing empty lines are removed. Multiple empty lines are
 // reduced to one, and trailing space on lines is trimmed. Unless the result
 // is empty, it is newline-terminated.
@@ -243,17 +243,10 @@ func (g *CommentGroup) Text() string {
 	for _, c := range comments {
 		// Remove comment markers.
 		// The parser has given us exactly the comment text.
-		switch c[1] {
-		case '/':
-			//-style comment (no newline at the end)
-			c = c[2:]
-			// strip first space - required for Example tests
-			if len(c) > 0 && c[0] == ' ' {
-				c = c[1:]
-			}
-		case '*':
-			/*-style comment */
-			c = c[2 : len(c)-2]
+		c = c[2:]
+		// strip first space - required for Example tests
+		if len(c) > 0 && c[0] == ' ' {
+			c = c[1:]
 		}
 
 		// Split on newlines.
@@ -394,7 +387,7 @@ type BottomLit struct {
 	expr
 }
 
-// An Ident node represents an left-hand side identifier,
+// An Ident node represents a left-hand side identifier,
 // including the underscore "_" identifier to represent top.
 type Ident struct {
 	NamePos token.Pos // identifier position
@@ -951,7 +944,7 @@ func (d *EmbedDecl) End() token.Pos { return d.Expr.End() }
 // ----------------------------------------------------------------------------
 // Files and packages
 
-// A File node represents a Go source file.
+// A File node represents a CUE source file.
 //
 // The Comments list contains all comments in the source file in order of
 // appearance, including the comments that are pointed to from other nodes
@@ -1006,6 +999,9 @@ func (f *File) PackageName() string {
 	for _, d := range f.Decls {
 		switch x := d.(type) {
 		case *Package:
+			if x.Name.Name == "_" {
+				return ""
+			}
 			return x.Name.Name
 		case *CommentGroup, *Attribute:
 		default:
