@@ -15,14 +15,15 @@
 package load
 
 import (
+	"cmp"
 	"fmt"
-	"os"
+	"io"
+	"io/fs"
 	pathpkg "path"
 	"path/filepath"
-	"sort"
+	"slices"
 	"strings"
 
-	"cuelang.org/go/cue/ast"
 	"cuelang.org/go/cue/build"
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/cue/token"
@@ -70,7 +71,7 @@ func (l *loader) importPkg(pos token.Pos, p *build.Instance) []*build.Instance {
 	defer l.stk.Pop()
 
 	cfg := l.cfg
-	ctxt := &cfg.fileSystem
+	ctxt := cfg.fileSystem
 
 	if p.Err != nil {
 		return []*build.Instance{p}
@@ -80,7 +81,6 @@ func (l *loader) importPkg(pos token.Pos, p *build.Instance) []*build.Instance {
 
 	if p.PkgName == "" {
 		if l.cfg.Package == "*" {
-			fp.ignoreOther = true
 			fp.allPackages = true
 			p.PkgName = "_"
 		} else {
@@ -131,39 +131,41 @@ func (l *loader) importPkg(pos token.Pos, p *build.Instance) []*build.Instance {
 	// have the same module scope and that there are no invalid modules.
 	inModule := false // if pkg == "_"
 	for _, d := range dirs {
-		if l.cfg.findRoot(d[1]) != "" {
+		if l.cfg.findModRoot(d[1]) != "" {
 			inModule = true
 			break
 		}
 	}
 
+	// Walk the parent directories up to the module root to add their files as well,
+	// since a package foo/bar/baz inherits from parent packages foo/bar and foo.
+	// See https://cuelang.org/docs/concept/modules-packages-instances/#instances.
 	for _, d := range dirs {
-		for dir := filepath.Clean(d[1]); ctxt.isDir(dir); {
-			files, err := ctxt.readDir(dir)
-			if err != nil && !os.IsNotExist(err) {
-				return retErr(errors.Wrapf(err, pos, "import failed reading dir %v", dirs[0][1]))
+		dir := filepath.Clean(d[1])
+		for {
+			sd, ok := l.dirCachedBuildFiles[dir]
+			if !ok {
+				sd = l.scanDir(dir)
+				l.dirCachedBuildFiles[dir] = sd
 			}
-			for _, f := range files {
-				if f.IsDir() {
-					continue
+			if err := sd.err; err != nil {
+				if errors.Is(err, fs.ErrNotExist) {
+					break
 				}
-				if f.Name() == "-" {
-					if _, err := cfg.fileSystem.stat("-"); !os.IsNotExist(err) {
-						continue
-					}
-				}
-				file, err := filetypes.ParseFile(f.Name(), filetypes.Input)
+				return retErr(errors.Wrapf(err, token.NoPos, "import failed reading dir %v", dir))
+			}
+			for _, name := range sd.filenames {
+				file, err := filetypes.ParseFileAndType(name, "", filetypes.Input)
 				if err != nil {
 					p.UnknownFiles = append(p.UnknownFiles, &build.File{
-						Filename:      f.Name(),
+						Filename:      name,
 						ExcludeReason: errors.Newf(token.NoPos, "unknown filetype"),
 					})
-					continue // skip unrecognized file types
+				} else {
+					fp.add(dir, file, 0)
 				}
-				fp.add(dir, file, importComment)
 			}
-
-			if p.PkgName == "" || !inModule || l.cfg.isRoot(dir) || dir == d[0] {
+			if p.PkgName == "" || !inModule || l.cfg.isModRoot(dir) || dir == d[0] {
 				break
 			}
 
@@ -187,9 +189,19 @@ func (l *loader) importPkg(pos token.Pos, p *build.Instance) []*build.Instance {
 		impPath, err := addImportQualifier(importPath(p.ImportPath), p.PkgName)
 		p.ImportPath = string(impPath)
 		if err != nil {
-			p.ReportError(err)
+			p.ReportError(errors.Promote(err, ""))
 		}
 
+		if len(p.BuildFiles) == 0 &&
+			len(p.IgnoredFiles) == 0 &&
+			len(p.OrphanedFiles) == 0 &&
+			len(p.InvalidFiles) == 0 &&
+			len(p.UnknownFiles) == 0 {
+			// The package has no files in it. This can happen
+			// when the default package added in newFileProcessor
+			// doesn't have any associated files.
+			continue
+		}
 		all = append(all, p)
 		rewriteFiles(p, cfg.ModuleRoot, false)
 		if errs := fp.finalize(p); errs != nil {
@@ -197,29 +209,92 @@ func (l *loader) importPkg(pos token.Pos, p *build.Instance) []*build.Instance {
 			return all
 		}
 
-		l.addFiles(cfg.ModuleRoot, p)
+		l.addFiles(p)
 		_ = p.Complete()
 	}
-	sort.Slice(all, func(i, j int) bool {
-		return all[i].Dir < all[j].Dir
+	slices.SortFunc(all, func(a, b *build.Instance) int {
+		// Instances may share the same directory but have different package names.
+		// Sort by directory first, then by package name.
+		if c := cmp.Compare(a.Dir, b.Dir); c != 0 {
+			return c
+		}
+
+		return cmp.Compare(a.PkgName, b.PkgName)
 	})
 	return all
 }
 
-// _loadFunc is the method used for the value of l.loadFunc.
+func (l *loader) scanDir(dir string) cachedDirFiles {
+	files, err := l.cfg.fileSystem.readDir(dir)
+	if err != nil {
+		return cachedDirFiles{
+			err: err,
+		}
+	}
+	filenames := make([]string, 0, len(files))
+	for _, f := range files {
+		if f.IsDir() {
+			continue
+		}
+		name := f.Name()
+		if name == "-" {
+			// The name "-" has a special significance to the file types
+			// logic, but only when specified directly on the command line.
+			// We don't want an actual file named "-" to have special
+			// significant, so avoid that by making sure we don't see a naked "-"
+			// even when a file named "-" is present in a directory.
+			name = "./-"
+		}
+		filenames = append(filenames, name)
+	}
+	return cachedDirFiles{
+		filenames: filenames,
+	}
+}
+
+func setFileSource(cfg *Config, f *build.File) error {
+	if f.Source != nil {
+		return nil
+	}
+	fullPath := f.Filename
+	if fullPath == "-" {
+		b, err := io.ReadAll(cfg.stdin())
+		if err != nil {
+			return errors.Newf(token.NoPos, "read stdin: %v", err)
+		}
+		f.Source = b
+		return nil
+	}
+	if !filepath.IsAbs(fullPath) {
+		fullPath = filepath.Join(cfg.Dir, fullPath)
+		// Ensure that encoding.NewDecoder will work correctly.
+		f.Filename = fullPath
+	}
+	if fi := cfg.fileSystem.getOverlay(fullPath); fi != nil {
+		if fi.file != nil {
+			f.Source = fi.file
+		} else {
+			f.Source = fi.contents
+		}
+	}
+	return nil
+}
+
+func (l *loader) loadFunc() build.LoadFunc {
+	if l.cfg.SkipImports {
+		return nil
+	}
+	return l._loadFunc
+}
+
 func (l *loader) _loadFunc(pos token.Pos, path string) *build.Instance {
 	impPath := importPath(path)
 	if isLocalImport(path) {
 		return l.cfg.newErrInstance(errors.Newf(pos, "relative import paths not allowed (%q)", path))
 	}
 
-	// is it a builtin?
-	if strings.IndexByte(strings.Split(path, "/")[0], '.') == -1 {
-		if l.cfg.StdRoot != "" {
-			p := l.newInstance(pos, impPath)
-			_ = l.importPkg(pos, p)
-			return p
-		}
+	if isStdlibPackage(path) {
+		// It looks like a builtin.
 		return nil
 	}
 
@@ -235,27 +310,35 @@ func (l *loader) newRelInstance(pos token.Pos, path, pkgName string) *build.Inst
 		panic(fmt.Errorf("non-relative import path %q passed to newRelInstance", path))
 	}
 
-	var err errors.Error
-	dir := path
-
-	p := l.cfg.Context.NewInstance(path, l.loadFunc)
+	p := l.cfg.Context.NewInstance(path, l.loadFunc())
 	p.PkgName = pkgName
 	p.DisplayPath = filepath.ToSlash(path)
 	// p.ImportPath = string(dir) // compute unique ID.
 	p.Root = l.cfg.ModuleRoot
 	p.Module = l.cfg.Module
 
-	dir = filepath.Join(l.cfg.Dir, filepath.FromSlash(path))
-
+	var err errors.Error
 	if path != cleanImport(path) {
 		err = errors.Append(err, l.errPkgf(nil,
 			"non-canonical import path: %q should be %q", path, pathpkg.Clean(path)))
 	}
 
-	if importPath, e := l.importPathFromAbsDir(fsPath(dir), path); e != nil {
+	dir := filepath.Join(l.cfg.Dir, filepath.FromSlash(path))
+	if pkgPath, e := importPathFromAbsDir(l.cfg, dir, path); e != nil {
 		// Detect later to keep error messages consistent.
 	} else {
-		p.ImportPath = string(importPath)
+		// Add package qualifier if the configuration requires it.
+		name := l.cfg.Package
+		switch name {
+		case "_", "*":
+			name = ""
+		}
+		pkgPath, e := addImportQualifier(pkgPath, name)
+		if e != nil {
+			// Detect later to keep error messages consistent.
+		} else {
+			p.ImportPath = string(pkgPath)
+		}
 	}
 
 	p.Dir = dir
@@ -272,53 +355,52 @@ func (l *loader) newRelInstance(pos token.Pos, path, pkgName string) *build.Inst
 	return p
 }
 
-func (l *loader) importPathFromAbsDir(absDir fsPath, key string) (importPath, errors.Error) {
-	if l.cfg.ModuleRoot == "" {
-		return "", errors.Newf(token.NoPos,
-			"cannot determine import path for %q (root undefined)", key)
+func importPathFromAbsDir(c *Config, absDir string, origPath string) (importPath, error) {
+	if c.ModuleRoot == "" {
+		return "", fmt.Errorf("cannot determine import path for %q (root undefined)", origPath)
 	}
 
-	dir := filepath.Clean(string(absDir))
-	if !strings.HasPrefix(dir, l.cfg.ModuleRoot) {
-		return "", errors.Newf(token.NoPos,
-			"cannot determine import path for %q (dir outside of root)", key)
+	dir := filepath.Clean(absDir)
+	if !strings.HasPrefix(dir, c.ModuleRoot) {
+		return "", fmt.Errorf("cannot determine import path for %q (dir outside of root)", origPath)
 	}
 
-	pkg := filepath.ToSlash(dir[len(l.cfg.ModuleRoot):])
+	pkg := filepath.ToSlash(dir[len(c.ModuleRoot):])
 	switch {
 	case strings.HasPrefix(pkg, "/cue.mod/"):
 		pkg = pkg[len("/cue.mod/"):]
 		if pkg == "" {
-			return "", errors.Newf(token.NoPos,
-				"invalid package %q (root of %s)", key, modDir)
+			return "", fmt.Errorf("invalid package %q (root of %s)", origPath, modDir)
 		}
 
-	case l.cfg.Module == "":
-		return "", errors.Newf(token.NoPos,
-			"cannot determine import path for %q (no module)", key)
+	case c.Module == "":
+		return "", fmt.Errorf("cannot determine import path for %q (no module)", origPath)
 	default:
-		pkg = l.cfg.Module + pkg
+		impPath := module.ParseImportPath(c.Module)
+		impPath.Path += pkg
+		impPath.Qualifier = ""
+		pkg = impPath.String()
 	}
-
-	name := l.cfg.Package
-	switch name {
-	case "_", "*":
-		name = ""
-	}
-
-	return addImportQualifier(importPath(pkg), name)
+	return importPath(pkg), nil
 }
 
 func (l *loader) newInstance(pos token.Pos, p importPath) *build.Instance {
-	dir, name, err := l.absDirFromImportPath(pos, p)
-	i := l.cfg.Context.NewInstance(dir, l.loadFunc)
+	dir, modPath, err := l.absDirFromImportPath(pos, p)
+	i := l.cfg.Context.NewInstance(dir, l.loadFunc())
+	i.Err = errors.Append(i.Err, err)
 	i.Dir = dir
-	i.PkgName = name
+
+	parts := module.ParseImportPath(string(p))
+	i.PkgName = parts.Qualifier
+	if i.PkgName == "" {
+		i.Err = errors.Append(i.Err, l.errPkgf([]token.Pos{pos}, "cannot determine package name for %q; set it explicitly with ':'", p))
+	} else if i.PkgName == "_" {
+		i.Err = errors.Append(i.Err, l.errPkgf([]token.Pos{pos}, "_ is not a valid import path qualifier in %q", p))
+	}
 	i.DisplayPath = string(p)
 	i.ImportPath = string(p)
 	i.Root = l.cfg.ModuleRoot
-	i.Module = l.cfg.Module
-	i.Err = errors.Append(i.Err, err)
+	i.Module = modPath
 
 	return i
 }
@@ -327,39 +409,47 @@ func (l *loader) newInstance(pos token.Pos, p importPath) *build.Instance {
 // and a package name. The root directory must be set.
 //
 // The returned directory may not exist.
-func (l *loader) absDirFromImportPath(pos token.Pos, p importPath) (absDir, name string, err errors.Error) {
+func (l *loader) absDirFromImportPath(pos token.Pos, p importPath) (dir string, modPath string, _ errors.Error) {
+	dir, modPath, err := l.absDirFromImportPath1(pos, p)
+	if err != nil {
+		// Any error trying to determine the package location
+		// is a PackageError.
+		return "", "", l.errPkgf([]token.Pos{pos}, "%s", err.Error())
+	}
+	return dir, modPath, nil
+}
+
+func (l *loader) absDirFromImportPath1(pos token.Pos, p importPath) (absDir string, modPath string, err error) {
+	if p == "" {
+		return "", "", fmt.Errorf("empty import path")
+	}
 	if l.cfg.ModuleRoot == "" {
-		return "", "", errors.Newf(pos, "cannot import %q (root undefined)", p)
+		return "", "", fmt.Errorf("cannot import %q (root undefined)", p)
+	}
+	if isStdlibPackage(string(p)) {
+		return "", "", fmt.Errorf("standard library import path %q cannot be imported as a CUE package", p)
 	}
-	origp := p
 	// Extract the package name.
 	parts := module.ParseImportPath(string(p))
-	name = parts.Qualifier
-	p = importPath(parts.Unqualified().String())
-	if name == "" {
-		err = errors.Newf(pos, "empty package name in import path %q", p)
-	} else if strings.IndexByte(name, '.') >= 0 {
-		err = errors.Newf(pos,
-			"cannot determine package name for %q (set explicitly with ':')", p)
-	} else if !ast.IsValidIdent(name) {
-		err = errors.Newf(pos,
-			"implied package identifier %q from import path %q is not valid", name, p)
-	}
+	unqualified := parts.Unqualified().String()
 	if l.cfg.Registry != nil {
 		if l.pkgs == nil {
-			return "", name, errors.Newf(pos, "imports are unavailable because there is no cue.mod/module.cue file")
+			return "", "", fmt.Errorf("imports are unavailable because there is no cue.mod/module.cue file")
 		}
 		// TODO predicate registry-aware lookup on module.cue-declared CUE version?
 
-		// Note: use the original form of the import path because
-		// that's the form passed to modpkgload.LoadPackages
+		// Note: use the canonical form of the import path because
+		// that's the form passed to [modpkgload.LoadPackages]
 		// and hence it's available by that name via Pkg.
-		pkg := l.pkgs.Pkg(string(origp))
+		pkg := l.pkgs.Pkg(parts.Canonical().String())
+		// TODO(mvdan): using "unqualified" for the errors below doesn't seem right,
+		// should we not be using either the original path or the canonical path?
+		// The unqualified import path should only be used for filepath.FromSlash further below.
 		if pkg == nil {
-			return "", name, errors.Newf(pos, "no dependency found for package %q", p)
+			return "", "", fmt.Errorf("no dependency found for package %q", unqualified)
 		}
 		if err := pkg.Error(); err != nil {
-			return "", name, errors.Newf(pos, "cannot find package %q: %v", p, err)
+			return "", "", fmt.Errorf("cannot find package %q: %v", unqualified, err)
 		}
 		if mv := pkg.Mod(); mv.IsLocal() {
 			// It's a local package that's present inside one or both of the gen, usr or pkg
@@ -370,31 +460,37 @@ func (l *loader) absDirFromImportPath(pos token.Pos, p importPath) (absDir, name
 		} else {
 			locs := pkg.Locations()
 			if len(locs) > 1 {
-				return "", "", errors.Newf(pos, "package %q unexpectedly found in multiple locations", p)
+				return "", "", fmt.Errorf("package %q unexpectedly found in multiple locations", unqualified)
+			}
+			if len(locs) == 0 {
+				return "", "", fmt.Errorf("no location found for package %q", unqualified)
 			}
 			var err error
 			absDir, err = absPathForSourceLoc(locs[0])
 			if err != nil {
-				return "", name, errors.Newf(pos, "cannot determine source directory for package %q: %v", p, err)
+				return "", "", fmt.Errorf("cannot determine source directory for package %q: %v", unqualified, err)
 			}
 		}
-		return absDir, name, nil
+		return absDir, pkg.Mod().Path(), nil
 	}
 
 	// Determine the directory without using the registry.
 
-	sub := filepath.FromSlash(string(p))
-	switch hasPrefix := strings.HasPrefix(string(p), l.cfg.Module); {
+	sub := filepath.FromSlash(unqualified)
+	switch hasPrefix := strings.HasPrefix(unqualified, l.cfg.Module); {
 	case hasPrefix && len(sub) == len(l.cfg.Module):
+		modPath = l.cfg.Module
 		absDir = l.cfg.ModuleRoot
 
-	case hasPrefix && p[len(l.cfg.Module)] == '/':
+	case hasPrefix && unqualified[len(l.cfg.Module)] == '/':
+		modPath = l.cfg.Module
 		absDir = filepath.Join(l.cfg.ModuleRoot, sub[len(l.cfg.Module)+1:])
 
 	default:
+		modPath = "local"
 		absDir = filepath.Join(GenPath(l.cfg.ModuleRoot), sub)
 	}
-	return absDir, name, err
+	return absDir, modPath, err
 }
 
 func absPathForSourceLoc(loc module.SourceLoc) (string, error) {
@@ -408,3 +504,14 @@ func absPathForSourceLoc(loc module.SourceLoc) (string, error) {
 	}
 	return filepath.Join(osPath, loc.Dir), nil
 }
+
+// isStdlibPackage reports whether pkgPath looks like
+// an import from the standard library.
+func isStdlibPackage(pkgPath string) bool {
+	firstElem, _, _ := strings.Cut(pkgPath, "/")
+	if firstElem == "" {
+		return false // absolute paths like "/foo/bar"
+	}
+	// Paths like ".foo/bar", "./foo/bar", or "foo.com/bar" are not standard library import paths.
+	return strings.IndexByte(firstElem, '.') == -1
+}
