@@ -90,7 +90,13 @@ func (n *nodeContext) scheduleConjunct(c Conjunct, id CloseInfo) {
 		}
 
 		if id.cc.src != n.node {
-			panic("inconsistent state: nodes differ")
+			// TODO(#3406): raise a panic again.
+			//		out: d & { d }
+			//		d: {
+			//			kind: "foo" | "bar"
+			//			{ kind: "foo" } | { kind: "bar" }
+			//		}
+			// panic("inconsistent state: nodes differ")
 		}
 	default:
 
@@ -113,7 +119,7 @@ func (n *nodeContext) scheduleConjunct(c Conjunct, id CloseInfo) {
 	env := c.Env
 
 	if id.cc.isDef {
-		n.node.Closed = true
+		n.node.ClosedRecursive = true
 	}
 
 	switch x := c.Elem().(type) {
@@ -175,7 +181,7 @@ func (n *nodeContext) scheduleConjunct(c Conjunct, id CloseInfo) {
 		n.unshare()
 
 		// At this point we known we have at least an empty list.
-		n.updateCyclicStatus(id)
+		n.updateCyclicStatusV3(id)
 
 		env := &Environment{
 			Up:     env,
@@ -185,6 +191,8 @@ func (n *nodeContext) scheduleConjunct(c Conjunct, id CloseInfo) {
 
 	case *DisjunctionExpr:
 		n.unshare()
+		id := id
+		id.setOptionalV3(n)
 
 		// TODO(perf): reuse envDisjunct values so that we can also reuse the
 		// disjunct slice.
@@ -228,7 +236,7 @@ func (n *nodeContext) scheduleConjunct(c Conjunct, id CloseInfo) {
 func (n *nodeContext) scheduleStruct(env *Environment,
 	s *StructLit,
 	ci CloseInfo) {
-	n.updateCyclicStatus(ci)
+	n.updateCyclicStatusV3(ci)
 
 	// NOTE: This is a crucial point in the code:
 	// Unification dereferencing happens here. The child nodes are set to
@@ -270,10 +278,14 @@ loop2:
 		case *Comprehension, Expr:
 			// No need to increment and decrement, as there will be at least
 			// one entry.
-			if _, ok := s.Src.(*ast.File); !ok {
+			if _, ok := s.Src.(*ast.File); !ok && s.Src != nil {
 				// If this is not a file, the struct indicates the scope/
 				// boundary at which closedness should apply. This is not true
 				// for files.
+				// We should also not spawn if this is a nested Comprehension,
+				// where the spawn is already done as it may lead to spurious
+				// field not allowed errors. We can detect this with a nil s.Src.
+				// TODO(evalv3): use a more principled detection mechanism.
 				// TODO: set this as a flag in StructLit so as to not have to
 				// do the somewhat dangerous cast here.
 				ci, _ = ci.spawnCloseContext(n.ctx, 0)
@@ -293,6 +305,11 @@ loop2:
 				n.aStruct = s
 				n.aStructID = ci
 			}
+			ci := ci
+			if x.ArcType == ArcOptional {
+				ci.setOptionalV3(n)
+			}
+
 			fc := MakeConjunct(childEnv, x, ci)
 			// fc.CloseInfo.cc = nil // TODO: should we add this?
 			n.insertArc(x.Label, x.ArcType, fc, ci, true)
@@ -303,6 +320,7 @@ loop2:
 
 		case *Comprehension:
 			ci, cc := ci.spawnCloseContext(n.ctx, closeEmbed)
+			cc.decl = x
 			cc.incDependent(n.ctx, DEFER, nil)
 			defer cc.decDependent(n.ctx, DEFER, nil)
 			n.insertComprehension(childEnv, x, ci)
@@ -322,6 +340,8 @@ loop2:
 			n.scheduleTask(handleDynamic, childEnv, x, ci)
 
 		case *BulkOptionalField:
+			ci := ci
+			ci.setOptionalV3(n)
 
 			// All do not depend on each other, so can be added at once.
 			n.scheduleTask(handlePatternConstraint, childEnv, x, ci)
@@ -329,6 +349,7 @@ loop2:
 		case Expr:
 			// TODO: perhaps special case scalar Values to avoid creating embedding.
 			ci, cc := ci.spawnCloseContext(n.ctx, closeEmbed)
+			cc.decl = x
 
 			// TODO: do we need to increment here?
 			cc.incDependent(n.ctx, DEFER, nil) // decrement deferred below
@@ -340,7 +361,7 @@ loop2:
 		}
 	}
 	if hasEllipsis {
-		ci.cc.hasEllipsis = true
+		ci.cc.isTotal = true
 	}
 	if !hasEmbed {
 		n.aStruct = s
@@ -531,26 +552,26 @@ func (n *nodeContext) addNotify2(v *Vertex, c CloseInfo) []receiver {
 // Literal conjuncts
 
 func (n *nodeContext) insertValueConjunct(env *Environment, v Value, id CloseInfo) {
-	n.updateCyclicStatus(id)
+	n.updateCyclicStatusV3(id)
 
 	ctx := n.ctx
 
 	switch x := v.(type) {
 	case *Vertex:
-		if m, ok := x.BaseValue.(*StructMarker); ok {
+		if x.ClosedNonRecursive {
+			n.node.ClosedNonRecursive = true
+			var cc *closeContext
+			id, cc = id.spawnCloseContext(n.ctx, 0)
+			cc.isClosedOnce = true
+
+			if v, ok := x.BaseValue.(*Vertex); ok {
+				n.insertValueConjunct(env, v, id)
+				return
+			}
+		}
+		if _, ok := x.BaseValue.(*StructMarker); ok {
 			n.aStruct = x
 			n.aStructID = id
-			if m.NeedClose {
-				// TODO: In the new evaluator this is used to mark a struct
-				// as closed in the debug output. Once the old evaluator is
-				// gone, we could simplify this.
-				id.IsClosed = true
-				if ctx.isDevVersion() {
-					var cc *closeContext
-					id, cc = id.spawnCloseContext(n.ctx, 0)
-					cc.isClosedOnce = true
-				}
-			}
 		}
 
 		if !x.IsData() {
@@ -565,7 +586,7 @@ func (n *nodeContext) insertValueConjunct(env *Environment, v Value, id CloseInf
 			panic(fmt.Sprintf("invalid type %T", x.BaseValue))
 
 		case *ListMarker:
-			n.updateCyclicStatus(id)
+			n.updateCyclicStatusV3(id)
 
 			// TODO: arguably we know now that the type _must_ be a list.
 			n.scheduleTask(handleListVertex, env, x, id)
@@ -609,6 +630,9 @@ func (n *nodeContext) insertValueConjunct(env *Environment, v Value, id CloseInf
 	case *Disjunction:
 		// TODO(perf): reuse envDisjunct values so that we can also reuse the
 		// disjunct slice.
+		id := id
+		id.setOptionalV3(n)
+
 		d := envDisjunct{
 			env:     env,
 			cloneID: id,
@@ -675,7 +699,7 @@ func (n *nodeContext) insertValueConjunct(env *Environment, v Value, id CloseInf
 			k := 0
 			match := false
 			for _, c := range n.checks {
-				if y, ok := c.(*BoundValue); ok {
+				if y, ok := c.x.(*BoundValue); ok {
 					switch z := SimplifyBounds(ctx, n.kind, x, y); {
 					case z == y:
 						match = true
@@ -688,15 +712,16 @@ func (n *nodeContext) insertValueConjunct(env *Environment, v Value, id CloseInf
 			}
 			n.checks = n.checks[:k]
 			if !match {
-				n.checks = append(n.checks, x)
+				n.checks = append(n.checks, MakeConjunct(env, x, id))
 			}
 			return
 		}
 
 	case Validator:
 		// This check serves as simplifier, but also to remove duplicates.
+		cx := MakeConjunct(env, x, id)
 		for i, y := range n.checks {
-			if b := SimplifyValidator(ctx, x, y); b != nil {
+			if b, ok := SimplifyValidator(ctx, cx, y); ok {
 				n.checks[i] = b
 				return
 			}
@@ -711,7 +736,18 @@ func (n *nodeContext) insertValueConjunct(env *Environment, v Value, id CloseInf
 		if kind&(ListKind|StructKind) != 0 {
 			id.cc.hasTop = true
 		}
-		n.checks = append(n.checks, x)
+		n.checks = append(n.checks, cx)
+
+		// We use set the type of the validator argument here to ensure that
+		// validation considers the ultimate value of embedded validators,
+		// rather than assuming that the struct in which an expression is
+		// embedded is always a struct.
+		// TODO(validatorType): get rid of setting n.hasTop here.
+		k := x.Kind()
+		if k == TopKind {
+			n.hasTop = true
+		}
+		n.updateNodeType(k, x, id)
 
 	case *Vertex:
 	// handled above.
