@@ -167,7 +167,9 @@ type Runtime interface {
 	// type if available.
 	LoadType(t reflect.Type) (src ast.Expr, expr Expr, ok bool)
 
-	Settings() (internal.EvaluatorVersion, cuedebug.Config)
+	// ConfigureOpCtx configures the [*OpContext] with details such as
+	// evaluator version, debug options etc.
+	ConfigureOpCtx(ctx *OpContext)
 }
 
 type Config struct {
@@ -180,22 +182,26 @@ func New(v *Vertex, cfg *Config) *OpContext {
 	if cfg.Runtime == nil {
 		panic("nil Runtime")
 	}
-	version, flags := cfg.Runtime.Settings()
+
 	ctx := &OpContext{
 		Runtime:     cfg.Runtime,
 		Format:      cfg.Format,
 		vertex:      v,
-		Version:     version,
-		Config:      flags,
 		taskContext: schedConfig,
 	}
+	cfg.Runtime.ConfigureOpCtx(ctx)
+	ctx.stats.EvalVersion = ctx.Version
 	if v != nil {
 		ctx.e = &Environment{Up: nil, Vertex: v}
 	}
 	return ctx
 }
 
+// See also: [unreachableForDev]
 func (c *OpContext) isDevVersion() bool {
+	if c.Version == internal.EvalVersionUnset {
+		panic("OpContext was not provided with an evaluator version")
+	}
 	return c.Version == internal.DevVersion
 }
 
@@ -208,7 +214,8 @@ type OpContext struct {
 	Format func(Runtime, Node) string
 
 	cuedebug.Config
-	Version internal.EvaluatorVersion // Copied from Runtime
+	Version  internal.EvaluatorVersion // Copied from Runtime
+	TopoSort bool                      // Copied from Runtime
 
 	taskContext
 
@@ -228,6 +235,12 @@ type OpContext struct {
 	// structural cycle errors.
 	vertex *Vertex
 
+	// list of vertices that need to be finalized.
+	// TODO: remove this again once we have a proper way of detecting references
+	// across optional boundaries in hasAncestorV3. We can probably do this
+	// with an optional depth counter.
+	toFinalize []*Vertex
+
 	// These fields are used associate scratch fields for computing closedness
 	// of a Vertex. These fields could have been included in StructInfo (like
 	// Tomabechi's unification algorithm), but we opted for an indirection to
@@ -262,6 +275,11 @@ type OpContext struct {
 	// enabled.
 	inConstraint int
 
+	// inDetached indicates that inline structs evaluated in the current context
+	// should never be shared. This is the case, for instance, with the source
+	// for the for clause in a comprehension.
+	inDetached int
+
 	// inValidator defines whether full evaluation need to be enforced, for
 	// instance when comparing against bottom.
 	inValidator int
@@ -382,7 +400,11 @@ func (c *OpContext) addErrf(code ErrorCode, pos token.Pos, msg string, args ...i
 }
 
 func (c *OpContext) addErr(code ErrorCode, err errors.Error) {
-	c.AddBottom(&Bottom{Code: code, Err: err})
+	c.AddBottom(&Bottom{
+		Code: code,
+		Err:  err,
+		Node: c.vertex,
+	})
 }
 
 // AddBottom records an error in OpContext.
@@ -393,7 +415,10 @@ func (c *OpContext) AddBottom(b *Bottom) {
 // AddErr records an error in OpContext. It returns errors collected so far.
 func (c *OpContext) AddErr(err errors.Error) *Bottom {
 	if err != nil {
-		c.AddBottom(&Bottom{Err: err})
+		c.AddBottom(&Bottom{
+			Err:  err,
+			Node: c.vertex,
+		})
 	}
 	return c.errs
 }
@@ -404,7 +429,12 @@ func (c *OpContext) NewErrf(format string, args ...interface{}) *Bottom {
 	// TODO: consider renaming ot NewBottomf: this is now confusing as we also
 	// have Newf.
 	err := c.Newf(format, args...)
-	return &Bottom{Src: c.src, Err: err, Code: EvalError}
+	return &Bottom{
+		Src:  c.src,
+		Err:  err,
+		Code: EvalError,
+		Node: c.vertex,
+	}
 }
 
 // AddErrf records an error in OpContext. It returns errors collected so far.
@@ -524,14 +554,23 @@ func (c *OpContext) Lookup(env *Environment, r Resolver) (*Vertex, *Bottom) {
 //
 // TODO(errors): return boolean instead: only the caller has enough information
 // to generate a proper error message.
-func (c *OpContext) Validate(check Validator, value Value) *Bottom {
+func (c *OpContext) Validate(check Conjunct, value Value) *Bottom {
 	// TODO: use a position stack to push both values.
-	saved := c.src
+
+	// TODO(evalv3): move to PushConjunct once the migration is complete.
+	// Using PushConjunct also saves and restores the error, which may be
+	// impactful, so we want to do this in a separate commit.
+	// saved := c.PushConjunct(check)
+
+	src := c.src
+	ci := c.ci
 	c.src = check.Source()
+	c.ci = check.CloseInfo
 
-	err := check.validate(c, value)
+	err := check.x.(Validator).validate(c, value)
 
-	c.src = saved
+	c.src = src
+	c.ci = ci
 
 	return err
 }
@@ -615,6 +654,7 @@ func (c *OpContext) Evaluate(env *Environment, x Expr) (result Value, complete b
 		val = &Bottom{
 			Code: IncompleteError,
 			Err:  c.Newf("UNANTICIPATED ERROR"),
+			Node: env.Vertex,
 		}
 
 	}
@@ -628,6 +668,20 @@ func (c *OpContext) Evaluate(env *Environment, x Expr) (result Value, complete b
 	return val, true
 }
 
+// EvaluateKeepState does an evaluate, but leaves any errors an cycle info
+// within the context.
+func (c *OpContext) EvaluateKeepState(x Expr) (result Value) {
+	src := c.src
+	c.src = x.Source()
+
+	result, ci := c.evalStateCI(x, final(partial, concreteKnown))
+
+	c.src = src
+	c.ci = ci
+
+	return result
+}
+
 func (c *OpContext) evaluateRec(v Conjunct, state combinedFlags) Value {
 	x := v.Expr()
 	s := c.PushConjunct(v)
@@ -639,6 +693,7 @@ func (c *OpContext) evaluateRec(v Conjunct, state combinedFlags) Value {
 		val = &Bottom{
 			Code: IncompleteError,
 			Err:  c.Newf("UNANTICIPATED ERROR"),
+			Node: c.vertex,
 		}
 	}
 	_ = c.PopState(s)
@@ -658,10 +713,18 @@ func (c *OpContext) value(x Expr, state combinedFlags) (result Value) {
 }
 
 func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
+	result, _ = c.evalStateCI(v, state)
+	return result
+}
+
+func (c *OpContext) evalStateCI(v Expr, state combinedFlags) (result Value, ci CloseInfo) {
 	savedSrc := c.src
 	c.src = v.Source()
 	err := c.errs
 	c.errs = nil
+	// Save the old CloseInfo and restore after evaluate to avoid detecting
+	// spurious cycles.
+	saved := c.ci
 
 	defer func() {
 		c.errs = CombineErrors(c.src, c.errs, err)
@@ -692,23 +755,29 @@ func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
 			result = c.errs
 		}
 		c.src = savedSrc
+
+		// TODO(evalv3): this c.ci should be passed to the caller who may need
+		// it to continue cycle detection for partially evaluated values.
+		// Either this or we must prove that this is covered by structural cycle
+		// detection.
+		c.ci = saved
 	}()
 
 	switch x := v.(type) {
 	case Value:
-		return x
+		return x, c.ci
 
 	case Evaluator:
 		v := x.evaluate(c, state)
-		return v
+		return v, c.ci
 
 	case Resolver:
 		arc := x.resolve(c, state)
 		if c.HasErr() {
-			return nil
+			return nil, c.ci
 		}
 		if arc == nil {
-			return nil
+			return nil, c.ci
 		}
 		// TODO(deref): what is the right level of dereferencing here?
 		// DerefValue seems to work too.
@@ -719,15 +788,16 @@ func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
 		// TODO: is this indirect necessary?
 		// arc = arc.Indirect()
 
-		// Save the old CloseInfo and restore after evaluate to avoid detecting
-		// spurious cycles.
-		saved := c.ci
 		n := arc.state
 		if c.isDevVersion() {
 			n = arc.getState(c)
-		}
-		if n != nil {
-			c.ci, _ = n.markCycle(arc, nil, x, c.ci)
+			if n != nil {
+				c.ci, _ = n.detectCycleV3(arc, nil, x, c.ci)
+			}
+		} else {
+			if n != nil {
+				c.ci, _ = n.markCycle(arc, nil, x, c.ci)
+			}
 		}
 		c.ci.Inline = true
 
@@ -753,8 +823,8 @@ func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
 						}
 						err := c.Newf("cycle with field %v", x)
 						b := &Bottom{Code: CycleError, Err: err}
-						v.setValue(c, v.status, b)
-						return b
+						s.setBaseValue(b)
+						return b, c.ci
 						// TODO: use this instead, as is usual for incomplete errors,
 						// and also move this block one scope up to also apply to
 						// defined arcs. In both cases, though, doing so results in
@@ -763,13 +833,13 @@ func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
 						// return nil
 					}
 					c.undefinedFieldError(v, IncompleteError)
-					return nil
+					return nil, c.ci
 				}
 			}
 		}
 		v := c.evaluate(arc, x, state)
-		c.ci = saved
-		return v
+
+		return v, c.ci
 
 	default:
 		// This can only happen, really, if v == nil, which is not allowed.
@@ -851,11 +921,15 @@ func (c *OpContext) unifyNode(v Expr, state combinedFlags) (result Value) {
 
 		if c.isDevVersion() {
 			if n := v.getState(c); n != nil {
+				// A lookup counts as new structure. See the commend in Section
+				// "Lookups in inline cycles" in cycle.go.
+				n.hasNonCycle = true
+
 				// Always yield to not get spurious errors.
 				n.process(arcTypeKnown, yield)
 			}
 		} else {
-			if v.isUndefined() || state.vertexStatus() > v.status {
+			if v.isUndefined() || state.vertexStatus() > v.Status() {
 				c.unify(v, state)
 			}
 		}
@@ -948,7 +1022,9 @@ func (c *OpContext) lookup(x *Vertex, pos token.Pos, l Feature, flags combinedFl
 			c.unify(a, deprecated(c, partial))
 		}
 
-		if a.IsConstraint() {
+		// TODO(refRequired): see comment in unify.go:Vertex.lookup near the
+		// namesake TODO.
+		if a.ArcType == ArcOptional {
 			code := IncompleteError
 			if hasCycle {
 				code = CycleError
@@ -959,6 +1035,7 @@ func (c *OpContext) lookup(x *Vertex, pos token.Pos, l Feature, flags combinedFl
 				Permanent: x.status >= conjuncts,
 				Err: c.NewPosf(pos,
 					"cannot reference optional field: %s", label),
+				Node: x,
 			})
 		}
 	} else {
@@ -1005,6 +1082,7 @@ func (c *OpContext) lookup(x *Vertex, pos token.Pos, l Feature, flags combinedFl
 			Code:      code,
 			Permanent: permanent,
 			Err:       err,
+			Node:      x,
 		})
 	}
 	return a
@@ -1131,6 +1209,11 @@ func (c *OpContext) RawElems(v Value) []*Vertex {
 }
 
 func (c *OpContext) list(v Value) *Vertex {
+	if v != nil {
+		if a, ok := c.getDefault(v); ok {
+			v = a
+		}
+	}
 	x, ok := v.(*Vertex)
 	if !ok || !x.IsList() {
 		c.typeError(v, ListKind)
