@@ -123,6 +123,7 @@ const (
 	taskWAITING // task is blocked on a property of an arc to hold
 	taskSUCCESS
 	taskFAILED
+	taskCANCELLED
 )
 
 type schedState uint8
@@ -175,6 +176,8 @@ func (s schedState) String() string {
 // runMode indicates how to proceed after a condition could not be met.
 type runMode uint8
 
+//go:generate go run golang.org/x/tools/cmd/stringer -type=runMode
+
 const (
 	// ignore indicates that the new evaluator should not do any processing.
 	// This is mostly used in the transition from old to new evaluator and
@@ -195,20 +198,6 @@ const (
 	finalize
 )
 
-func (r runMode) String() string {
-	switch r {
-	case ignore:
-		return "ignore"
-	case attemptOnly:
-		return "attemptOnly"
-	case yield:
-		return "yield"
-	case finalize:
-		return "finalize"
-	}
-	return "unknown"
-}
-
 // condition is a bit mask of states that a task may depend on.
 //
 // There are generally two types of states: states that are met if all tasks
@@ -405,9 +394,6 @@ processNextTask:
 		}
 
 		switch {
-		case t.defunct:
-			continue
-
 		case t.state == taskRUNNING:
 			// TODO: we could store the current referring node that caused
 			// the cycle and then proceed up the stack to mark all tasks
@@ -671,6 +657,12 @@ func (s *scheduler) insertTask(t *task) {
 
 func runTask(t *task, mode runMode) {
 	if t.defunct {
+		if t.state != taskCANCELLED {
+			t.state = taskCANCELLED
+			if t.id.cc != nil {
+				t.id.cc.decDependent(t.node.ctx, TASK, nil)
+			}
+		}
 		return
 	}
 	t.node.Logf("============ RUNTASK %v %v", t.run.name, t.x)
@@ -711,7 +703,10 @@ func runTask(t *task, mode runMode) {
 	defer ctx.popTask()
 	if t.env != nil {
 		id := t.id
-		id.cc = nil // this is done to avoid struct args from passing fields up.
+		// This is done to avoid struct args from passing fields up.
+		// Use [task.updateCI] to get the current CloseInfo with this field
+		// restored.
+		id.cc = nil
 		s := ctx.PushConjunct(MakeConjunct(t.env, t.x, id))
 		defer ctx.PopState(s)
 	}
@@ -744,6 +739,13 @@ func runTask(t *task, mode runMode) {
 	}
 }
 
+// updateCI stitches back the closeContext that more removed from the CloseInfo
+// before in the given CloseInfo.
+func (t *task) updateCI(ci CloseInfo) CloseInfo {
+	ci.cc = t.id.cc
+	return ci
+}
+
 // waitFor blocks task t until the needs for scheduler s are met.
 func (t *task) waitFor(s *scheduler, needs condition) {
 	if s.meets(needs) {
