@@ -77,12 +77,12 @@ func (n *nodeContext) initBare() {
 	n.blockOn(scalarKnown | listTypeKnown | arcTypeKnown)
 
 	if v.Label.IsDef() {
-		v.Closed = true
+		v.ClosedRecursive = true
 	}
 
 	if v.Parent != nil {
-		if v.Parent.Closed {
-			v.Closed = true
+		if v.Parent.ClosedRecursive {
+			v.ClosedRecursive = true
 		}
 	}
 }
@@ -119,6 +119,7 @@ func (n *nodeContext) scheduleConjuncts() {
 	root.decDependent(ctx, INIT, nil)
 }
 
+// TODO(evalv3): consider not returning a result at all.
 func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 	if c.LogEval > 0 {
 		c.nest++
@@ -129,10 +130,32 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 		}()
 	}
 
+	if c.evalDepth == 0 {
+		defer func() {
+			// This loop processes nodes that need to be evaluated, but should be
+			// evaluated outside of the stack to avoid structural cycle detection.
+			// See comment at toFinalize.
+			a := c.toFinalize
+			c.toFinalize = c.toFinalize[:0]
+			for _, x := range a {
+				x.Finalize(c)
+			}
+		}()
+	}
+
 	if mode == ignore {
 		return false
 	}
 
+	// Note that the state of a node can be removed before the node is.
+	// This happens with the close builtin, for instance.
+	// See TestFromAPI in pkg export.
+	// TODO(evalv3): find something more principled.
+	if v.state == nil && v.cc() != nil && v.cc().conjunctCount == 0 {
+		v.status = finalized
+		return true
+	}
+
 	n := v.getState(c)
 	if n == nil {
 		return true // already completed
@@ -152,7 +175,7 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 	// Note that if mode is final, we will guarantee that the conditions for
 	// this if clause are met down the line. So we assume this is already the
 	// case and set the signal accordingly if so.
-	if v.Label.IsLet() || v.IsDynamic || v.Parent.allChildConjunctsKnown() || mode == finalize {
+	if !v.Rooted() || v.Parent.allChildConjunctsKnown() || mode == finalize {
 		n.signal(allAncestorsProcessed)
 	}
 
@@ -164,7 +187,7 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 	w := v.DerefDisjunct()
 	if w != v {
 		// Should resolve with dereference.
-		v.Closed = w.Closed
+		v.ClosedRecursive = w.ClosedRecursive
 		v.status = w.status
 		v.ChildErrors = CombineErrors(nil, v.ChildErrors, w.ChildErrors)
 		v.Arcs = nil
@@ -209,6 +232,16 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 		}
 	}
 
+	if v, ok := n.node.BaseValue.(*Vertex); ok && n.sharedID.CycleType == NoCycle {
+		if n.ctx.hasDepthCycle(v) {
+			n.reportCycleError()
+			return true
+		}
+		// We unify here to proactively detect cycles. We do not need to,
+		// nor should we, if have have already found one.
+		v.unify(n.ctx, needs, mode)
+	}
+
 	// At this point, no more conjuncts will be added, so we could decrement
 	// the notification counters.
 
@@ -222,7 +255,7 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 		}
 
 		switch {
-		case assertStructuralCycle(n):
+		case assertStructuralCycleV3(n):
 		// TODO: consider bailing on error if n.errs != nil.
 		case n.completeAllArcs(needs, mode):
 		}
@@ -252,7 +285,7 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 		}
 
 	case needs&fieldSetKnown != 0:
-		n.evalArcTypes()
+		n.evalArcTypes(mode)
 	}
 
 	if err := n.getErr(); err != nil {
@@ -260,7 +293,7 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 		if b := n.node.Bottom(); b != nil {
 			err = CombineErrors(nil, b, err)
 		}
-		n.node.BaseValue = err
+		n.setBaseValue(err)
 	}
 
 	if mode == attemptOnly {
@@ -281,18 +314,18 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 		v.ChildErrors = nil
 		v.Arcs = nil
 
-		result := w.unify(c, needs, mode)
-
 		// Set control fields that are referenced without dereferencing.
-		if w.Closed {
-			v.Closed = true
+		if w.ClosedRecursive {
+			v.ClosedRecursive = true
 		}
+		// NOTE: setting ClosedNonRecursive is not necessary, as it is
+		// handled by scheduleValue.
 		if w.HasEllipsis {
 			v.HasEllipsis = true
 		}
 		v.status = w.status
 
-		return result
+		return true
 	}
 
 	// TODO: adding this is wrong, but it should not cause the snippet below
@@ -310,23 +343,38 @@ func (v *Vertex) unify(c *OpContext, needs condition, mode runMode) bool {
 
 	// validationCompleted
 	if n.completed&(subFieldsProcessed) != 0 {
-		n.node.HasEllipsis = n.node.cc.hasEllipsis
+		n.node.HasEllipsis = n.node.cc().isTotal
+
+		// The next piece of code used to address the following case
+		// (order matters)
+		//
+		// 		c1: c: [string]: f2
+		// 		f2: c1
+		// 		Also: cycle/issue990
+		//
+		// However, with recent changes, it no longer matters. Simultaneously,
+		// this causes a hang in the following case:
+		//
+		// 		_self: x: [...and(x)]
+		// 		_self
+		// 		x: [1]
+		//
+		// For this reason we disable it now. It may be the case that we need
+		// to enable it for computing disjunctions.
+		//
+		n.incDepth()
+		defer n.decDepth()
+
+		if pc := n.node.PatternConstraints; pc != nil {
+			for _, c := range pc.Pairs {
+				c.Constraint.unify(n.ctx, allKnown, attemptOnly)
+			}
+		}
 
 		n.node.updateStatus(finalized)
 
 		defer n.unmarkOptional(n.markOptional())
 
-		// The next piece of code addresses the following case.
-		// order matters
-		// c1: c: [string]: f2
-		// f2: c1
-		// Also: cycle/issue990
-		if pc := n.node.PatternConstraints; pc != nil {
-			for _, c := range pc.Pairs {
-				c.Constraint.Finalize(n.ctx)
-			}
-		}
-
 		if DebugDeps {
 			RecordDebugGraph(n.ctx, n.node, "Finalize")
 		}
@@ -398,7 +446,7 @@ func (n *nodeContext) completeNodeTasks(mode runMode) {
 	// - pending arcs (or incoming COMPS)
 	// TODO: replace with something more principled that does not piggyback on
 	// debug information.
-	for _, r := range v.cc.externalDeps {
+	for _, r := range v.cc().externalDeps {
 		src := r.src
 		a := &src.arcs[r.index]
 		if a.decremented || a.kind != NOTIFY {
@@ -426,37 +474,18 @@ func (n *nodeContext) completeNodeTasks(mode runMode) {
 
 		cc.decDependent(n.ctx, ROOT, nil) // REF(decrement:nodeDone)
 	}
-
-	return
 }
 
 func (n *nodeContext) updateScalar() {
 	// Set BaseValue to scalar, but only if it was not set before. Most notably,
 	// errors should not be discarded.
 	if n.scalar != nil && (!n.node.IsErr() || isCyclePlaceholder(n.node.BaseValue)) {
-		n.node.BaseValue = n.scalar
+		n.setBaseValue(n.scalar)
 		n.signal(scalarKnown)
 	}
 }
 
 func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
-	if n.node.status == evaluatingArcs {
-		// NOTE: this was an "incomplete" error pre v0.6. If this is a problem
-		// we could make this a CycleError. Technically, this may be correct,
-		// as it is possible to make the values exactly as the inserted
-		// values. It seems more user friendly to just disallow this, though.
-		// TODO: make uniform error messages
-		// see compbottom2.cue:
-		n.ctx.addErrf(CycleError, pos(n.node), "mutual dependency")
-		n.node.IsCyclic = true
-		// Consider using this, although not all
-		// mutual dependencies are irrecoverable.
-		// n.reportCycleError()
-	}
-
-	// TODO: remove the use of updateStatus as a cycle detection mechanism.
-	n.node.updateStatus(evaluatingArcs)
-
 	if n.underlying != nil {
 		// References within the disjunct may end up referencing the layer that
 		// this node overlays. Also for these nodes we want to be able to detect
@@ -476,7 +505,7 @@ func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
 	// Investigate how to work around this.
 	n.completeNodeTasks(finalize)
 
-	for _, r := range n.node.cc.externalDeps {
+	for _, r := range n.node.cc().externalDeps {
 		src := r.src
 		a := &src.arcs[r.index]
 		if a.decremented {
@@ -493,13 +522,14 @@ func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
 	}
 
 	n.incDepth()
+	defer n.decDepth()
 
 	// XXX(0.7): only set success if needs complete arcs.
 	success := true
-	// Visit arcs recursively to validate and compute error.
-	for n.arcPos < len(n.node.Arcs) {
-		a := n.node.Arcs[n.arcPos]
-		n.arcPos++
+	// Visit arcs recursively to validate and compute error. Use index instead
+	// of range in case the Arcs grows during processing.
+	for arcPos := 0; arcPos < len(n.node.Arcs); arcPos++ {
+		a := n.node.Arcs[arcPos]
 
 		if !a.unify(n.ctx, needs, mode) {
 			success = false
@@ -521,7 +551,7 @@ func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
 		if !a.Label.IsLet() && a.ArcType <= ArcRequired {
 			a := a.DerefValue()
 			if err := a.Bottom(); err != nil {
-				n.node.AddChildError(err)
+				n.AddChildError(err)
 			}
 			success = true // other arcs are irrelevant
 		}
@@ -552,8 +582,6 @@ func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
 		}
 	}
 
-	n.decDepth()
-
 	k := 0
 	for _, a := range n.node.Arcs {
 		if a.ArcType != ArcNotPresent {
@@ -563,6 +591,32 @@ func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
 	}
 	n.node.Arcs = n.node.Arcs[:k]
 
+	// TODO: perhaps this code can go once we have builtins for comparing to
+	// bottom.
+	for _, c := range n.postChecks {
+		ctx := n.ctx
+		f := ctx.PushState(c.env, c.expr.Source())
+
+		v := ctx.evalState(c.expr, oldOnly(finalized))
+		v, _ = ctx.getDefault(v)
+		v = Unwrap(v)
+
+		switch _, isError := v.(*Bottom); {
+		case isError == c.expectError:
+		default:
+			n.node.AddErr(ctx, &Bottom{
+				Src:  c.expr.Source(),
+				Code: CycleError,
+				Node: n.node,
+				Err: ctx.NewPosf(pos(c.expr),
+					"circular dependency in evaluation of conditionals: %v changed after evaluation",
+					ctx.Str(c.expr)),
+			})
+		}
+
+		ctx.PopState(f)
+	}
+
 	// This should be called after all arcs have been processed, because
 	// whether sharing is possible or not may depend on how arcs with type
 	// ArcPending will resolve.
@@ -596,12 +650,12 @@ func (n *nodeContext) completeAllArcs(needs condition, mode runMode) bool {
 	return success
 }
 
-func (n *nodeContext) evalArcTypes() {
+func (n *nodeContext) evalArcTypes(mode runMode) {
 	for _, a := range n.node.Arcs {
 		if a.ArcType != ArcPending {
 			continue
 		}
-		a.unify(n.ctx, arcTypeKnown, yield)
+		a.unify(n.ctx, arcTypeKnown, mode)
 		// Ensure the arc is processed up to the desired level
 		if a.ArcType == ArcPending {
 			// TODO: cancel tasks?
@@ -610,6 +664,13 @@ func (n *nodeContext) evalArcTypes() {
 	}
 }
 
+func root(v *Vertex) *Vertex {
+	for v.Parent != nil {
+		v = v.Parent
+	}
+	return v
+}
+
 func (v *Vertex) lookup(c *OpContext, pos token.Pos, f Feature, flags combinedFlags) *Vertex {
 	task := c.current()
 	needs := flags.conditions()
@@ -626,10 +687,19 @@ func (v *Vertex) lookup(c *OpContext, pos token.Pos, f Feature, flags combinedFl
 		// proceed with partial data, in which case a "pending" arc will be
 		// created to be completed later.
 
-		// Report error for now.
+		// Propagate error if the error is from a different package. This
+		// compensates for the fact that we do not fully evaluate the package.
 		if state.hasErr() {
-			c.AddBottom(state.getErr())
+			err := state.getErr()
+			if err != nil && err.Node != nil && root(err.Node) != root(v) {
+				c.AddBottom(err)
+			}
 		}
+
+		// A lookup counts as new structure. See the commend in Section
+		// "Lookups in inline cycles" in cycle.go.
+		state.hasNonCycle = true
+
 		// TODO: ideally this should not be run at this point. Consider under
 		// which circumstances this is still necessary, and at least ensure
 		// this will not be run if node v currently has a running task.
@@ -687,7 +757,7 @@ func (v *Vertex) lookup(c *OpContext, pos token.Pos, f Feature, flags combinedFl
 				break
 			}
 		}
-		arcState.completeNodeTasks(attemptOnly)
+		arcState.completeNodeTasks(yield)
 
 		// Child nodes, if pending and derived from a comprehension, may
 		// still cause this arc to become not pending.
@@ -701,11 +771,18 @@ func (v *Vertex) lookup(c *OpContext, pos token.Pos, f Feature, flags combinedFl
 
 		switch runMode {
 		case ignore, attemptOnly:
+			// TODO(cycle): ideally, we should be able to require that the
+			// arcType be known at this point, but that does not seem to work.
+			// Revisit once we have the structural cycle detection in place.
+
 			// TODO: should we avoid notifying ArcPending vertices here?
 			if task != nil {
 				arcState.addNotify2(task.node.node, task.id)
 			}
-			return arcReturn
+			if arc.ArcType == ArcPending {
+				return arcReturn
+			}
+			goto handleArcType
 
 		case yield:
 			arcState.process(needs, yield)
@@ -718,14 +795,25 @@ func (v *Vertex) lookup(c *OpContext, pos token.Pos, f Feature, flags combinedFl
 		}
 	}
 
+handleArcType:
 	switch arc.ArcType {
-	case ArcMember:
+	case ArcMember, ArcRequired:
 		return arcReturn
 
-	case ArcOptional, ArcRequired:
+	case ArcOptional:
+		// Technically, this failure also applies to required fields. We assume
+		// however, that if a reference field that is made regular will already
+		// result in an error, so that piling up another error is not strictly
+		// necessary. Note that the spec allows for eliding an error if it is
+		// guaranteed another error is generated elsewhere. This does not
+		// properly cover the case where a reference is made directly within the
+		// definition, but this is fine for the purpose it serves.
+		// TODO(refRequired): revisit whether referencing required fields should
+		// fail.
 		label := f.SelectorString(c.Runtime)
 		b := &Bottom{
 			Code: IncompleteError,
+			Node: v,
 			Err: c.NewPosf(pos,
 				"cannot reference optional field: %s", label),
 		}
